class FoxtrotError(RuntimeError):
  def __init__(self, message):
    super(FoxtrotError, self).__init__(message)

class ParameterError(FoxtrotError):
  def __init__(self, param):
    self.parameter = param
    super(ParameterError, self).__init__("Missing parameter: {}".format(param))

class APIResponseError(FoxtrotError):
  pass

class APITimeoutError(FoxtrotError):
  def __init__(self, count):
    self.count = count
    super(APITimeoutError, self).__init__("Request timed out after {} seconds.".format(count))



from setuptools import setup

execfile('foxtrot/version.py')

def readme():
  # Generated with `pandoc --from=markdown --to=rst --output=README.rst README.md`
  with open('README.rst') as f:
    return f.read()

setup(
  name = 'foxtrot-client',
  packages = ['foxtrot'],
  version = VERSION,
  description = 'Foxtrot Client Library',
  long_description = readme(),
  author = 'Yasyf Mohamedali',
  author_email = 'yasyf@foxtrot.io',
  url = 'https://github.com/FoxtrotSystems/api-client-python',
  download_url = 'https://github.com/FoxtrotSystems/api-client-python/tarball/' + VERSION,
  license = 'MIT',
  keywords = ['foxtrot', 'route optimization'],
  install_requires = ['requests'],
  classifiers= [
    'License :: OSI Approved :: MIT License',
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.6',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.2',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4'
  ],
  use_2to3 = True
)



import abc, time
from errors import APIResponseError, APITimeoutError

class FoxtrotResponse(object):

  __metaclass__  = abc.ABCMeta

  def __init__(self, endpoint, resp, api_instance):
    self.endpoint = endpoint
    self.resp = resp
    self.api_instance = api_instance

  @abc.abstractmethod
  def is_complete(self):
    "True if the server is finished processing your request."
    raise NotImplementedError()

  @abc.abstractmethod
  def poll(self):
    "Queries the server for the status of your request, returning a FoxtrotResponse object."
    raise NotImplementedError()

  @abc.abstractmethod
  def get_result(self):
    "Returns the result from the server, if the server is finished processing your request."
    raise NotImplementedError()

  def poll_and_block(self, interval=1, timeout=60):
    "Blocks until the server is finished processing your request, up to a timeout limit."
    count = 0
    instance = self
    while not instance.is_complete():
      instance = instance.poll()
      time.sleep(interval)
      count += interval
      if count >= timeout:
        raise APITimeoutError(count)
    return instance

  @staticmethod
  def response_for(endpoint, resp, api_instance):
    if resp['status'] == 'pending':
      return PendingRespose(endpoint, resp, api_instance)
    elif resp['status'] == 'success':
      return SuccessResponse(endpoint, resp, api_instance)
    else:
      return ErrorResponse(endpoint, resp, api_instance)

class ErrorResponse(FoxtrotResponse):
  def __init__(self, endpoint, resp, api_instance):
    super(ErrorResponse, self).__init__(endpoint, resp, api_instance)
    self.message = resp.get('message')
    self.exception = APIResponseError(self.message)
    raise self.exception

  def is_complete(self):
    return True

  def poll(self):
    raise self.exception

  def get_result(self):
    return None

  def __repr__(self):
    return "ErrorResponse ({})".format(self.message)

class PendingRespose(FoxtrotResponse):
  def __init__(self, endpoint, resp, api_instance):
    super(PendingRespose, self).__init__(endpoint, resp, api_instance)
    self.txid = resp['txid']

  def is_complete(self):
    return False

  def poll(self):
    return self.api_instance.poll(self.txid)

  def get_result(self):
    return None

  def __repr__(self):
    return "PendingRespose ({})".format(self.txid)

class SuccessResponse(FoxtrotResponse):
  def __init__(self, endpoint, resp, api_instance):
    super(SuccessResponse, self).__init__(endpoint, resp, api_instance)
    self.txid = resp['txid']
    self.__data = resp['response']

  def is_complete(self):
    return True

  def poll(self):
    return self

  def get_result(self):
    return self.__data

  def __repr__(self):
    return "SuccessResponse ({})".format(self.txid)



import requests, json, re
from urllib import urlencode
from response import FoxtrotResponse
from errors import ParameterError
import types

class Foxtrot(object):
  """Foxtrot API Client"""

  host = 'api.foxtrot.io'
  ssl = True
  api_root = ''

  def __init__(self, api_key, version='v1'):
    self.api_key = api_key
    self.version = version
    self.headers = {'content-type': 'application/json'}
    self.__set_methods(self.__fetch_methods())

  def _proto(self):
    return "https" if self.ssl else "http"

  def _build_url(self, endpoint, args):
    args['key'] = self.api_key
    query = urlencode(args)
    return "{0}://{1}{2}/{3}/{4}?{5}".format(self._proto(), self.host, self.api_root, self.version, endpoint, query)

  def __make_request_no_data(self, method, endpoint, args):
    url = self._build_url(endpoint, args)
    response = getattr(requests, method)(url, headers=self.headers)
    return response.json()

  def __make_request_data(self, method, endpoint, data, args):
    url = self._build_url(endpoint, args)
    response = getattr(requests, method)(url, data=json.dumps(data), headers=self.headers)
    return response.json()

  def get(self, endpoint, data, args):
    args.update(data)
    return self.__make_request_no_data('get', endpoint, args)

  def delete(self, endpoint, data, args):
    args.update(data)
    return self.__make_request_no_data('delete', endpoint, args)

  def post(self, endpoint, data, args):
    return self.__make_request_data('post', endpoint, data, args)

  def put(self, endpoint, data, args):
    return self.__make_request_data('put', endpoint, data, args)

  def patch(self, endpoint, data, args):
    return self.__make_request_data('patch', endpoint, data, args)

  def poll(self, txid):
    return self.get_poll(txid=txid)

  def __fetch_methods(self):
    resp = requests.get("{0}://{1}{2}".format(self._proto(), self.host, self.api_root))
    return resp.json()['endpoints'][self.version]

  def __create_method(self, endpoint, endpoint_name, endpoint_method_name, http_method):
    def f(self, *args, **kwargs):
      for param in endpoint['params'][http_method]:
        if param not in kwargs:
          raise ParameterError(param)
      query = kwargs.pop('_query', {})
      resp = getattr(self, http_method.lower())(endpoint_name, kwargs, query)
      return FoxtrotResponse.response_for(endpoint_name, resp, self)
    f.__name__ = str(endpoint_method_name)
    f.__doc__ = endpoint['description'][http_method]
    return f

  def __set_method(self, endpoint):
    for http_method in endpoint['methods']:
      endpoint_name = (endpoint['path']
                      .replace('/{}/'.format(self.version), '', 1)
                      .replace('/', '_'))
      endpoint_name = re.sub(r'_<(.*)>', '', endpoint_name)
      endpoint_method_name = http_method.lower() + '_' + endpoint_name
      method = types.MethodType(self.__create_method(endpoint, endpoint_name,
                endpoint_method_name, http_method), self)
      setattr(self, endpoint_method_name, method)

  def __set_methods(self, endpoints):
    for endpoint in endpoints.values():
      if 'url' in endpoint:
        self.__set_method(endpoint)
      else:
        self.__set_methods(endpoint)
